#
#  keep_best_hit_when_same_query.pl
#
#  Version 0.1.1 (July 5, 2012)
#
#  Copyright (c) 2012 Kirill Kryukov
#
#  This software is provided 'as-is', without any express or implied
#  warranty. In no event will the authors be held liable for any damages
#  arising from the use of this software.
#
#  Permission is granted to anyone to use this software for any purpose,
#  including commercial applications, and to alter it and redistribute it
#  freely, subject to the following restrictions:
#
#  1. The origin of this software must not be misrepresented; you must not
#     claim that you wrote the original software. If you use this software
#     in a product, an acknowledgment in the product documentation would be
#     appreciated but is not required.
#  2. Altered source versions must be plainly marked as such, and must not be
#     misrepresented as being the original software.
#  3. This notice may not be removed or altered from any source distribution.
#

use strict;



my @number_to_alphanumeric_code = (
    '0','1','2','3','4','5','6','7','8','9',
    'A','B','C','D','E','F','G','H','I','J',
    'K','L','M','N','O','P','Q','R','S','T',
    'U','V','W','X','Y','Z','a','b','c','d',
    'e','f','g','h','i','j','k','l','m','n',
    'o','p','q','r','s','t','u','v','w','x',
    'y','z'
);

my @alphanumeric_code_to_number = (-1) x 256;

for (my $i=0; $i<62; $i++)
{
    $alphanumeric_code_to_number[ord($number_to_alphanumeric_code[$i])] = $i;
}



my ($qid,$sid,$best_sid,$best_score,$best_e,$score,$e,$nparts) = ('','','','','','','',0);
my @best_hits = ();
my @hits = ();
my ($n_processed,$n_kept) = (0,0);



while (<STDIN>)
{
    chomp;

    if (/^#/)
    {
        print_hit();
        print $_, "\n";
        next;
    }

    if (/^\s{2}/)
    {
        print_hit();
        if (!/^\s{2}(\S+)$/) { die "Can't parse query index \"$_\"\n"; }
        $qid = $1;
        next;
    }

    if ($qid eq '') { die "Hit with unknown query \"$_\"\n"; }

    if (/^\s/)
    {
        if (scalar(@hits) != $nparts) { die "Incomplete set of hits found\n"; }
        ($score,$e,$nparts) = ('','',0);
        @hits = ();
        if (!/^\s(\S+)$/) { die "Can't parse subject index \"$_\"\n"; }
        $sid = $1;
        next;
    }

    if ($sid eq '') { die "Hit with unknown subject: \"$_\"\n"; }

    if (!/^\S+\s\S+\s\S*\s\S*\s(\S+)\s(\S+)\s(\S*)/) { die "Can't parse hit \"$_\"\n"; }
    my ($score_code,$e_code,$nparts_code) = ($1,$2,$3);

    my $score1 = decode_scientific($score_code);
    my $e1 = decode_scientific($e_code);
    my $nparts1 = ($nparts_code eq '') ? 1 : decode_integer_alphanumeric($nparts_code);

    if (@hits)
    {
        if ($nparts1 != $nparts) { die "nparts of a multi-part hit does not match\n"; }
        if ($e1 ne $e) { die "evalue of a multi-part hit does not match\n"; }
    }
    else
    {
        if ($nparts != 0) { die "Incomplete set of hits found\n"; }
        $nparts = $nparts1;
        $e = $e1;
    }

    $score += $score1;
    push @hits, $_;

    if (scalar(@hits) == $nparts)
    {
        $n_processed++;

        if (!@best_hits or $e < $best_e or ($e == $best_e and $score > $best_score))
        {
            $best_sid = $sid;
            $best_score = $score;
            $best_e = $e;
            @best_hits = @hits;
        }

        ($score,$e,$nparts) = ('','',0);
        @hits = ();
    }
}
print_hit();
print STDERR "Done, $n_kept hits kept out of $n_processed\n";



sub print_hit
{
    if (scalar(@hits) != $nparts) { die "Incomplete set of hits found\n"; }

    if ($qid ne '')
    {
        print '  ', $qid, "\n";
        if ($best_sid ne '' and @best_hits)
        {
            print ' ', $best_sid, "\n";
            foreach my $hit (@best_hits)
            {
                print $hit, "\n";
            }
            $n_kept++;
        }
    }

    ($qid,$sid,$best_sid,$best_score,$best_e,$score,$e,$nparts) = ('','','','','','','',0);
    @best_hits = ();
    @hits = ();
}



sub decode_integer_base94
{
    my ($s) = @_;
    if ($s eq '') { return undef; }
    my $n = 0;
    my $l = length($s);
    for (my $i=0; $i<$l; $i++)
    {
        $n = $n * 94 + ord(substr($s,$i,1)) - 33;
    }
    return $n;
}



sub decode_integer_alphanumeric
{
    my ($s) = @_;
    if ($s eq '') { return undef; }
    my $n = 0;
    my $l = length($s);
    for (my $i=0; $i<$l; $i++)
    {
        $n = $n * 62 + $alphanumeric_code_to_number[ord(substr($s,$i,1))];
    }
    return $n;
}



sub decode_float
{
    my ($code) = @_;
    my $di = index($code,'.');
    if ($di >= 0)
    {
        return decode_integer_alphanumeric(substr($code,0,$di)) . '.' . reverse(decode_integer_alphanumeric(substr($code,$di+1)));
    }
    else
    {
        return decode_integer_alphanumeric($code);
    }
}



sub decode_scientific
{
    my ($code) = @_;
    my $pi = index($code,'+');
    my $mi = index($code,'-');
    if ($pi >= 0 or $mi >= 0)
    {
        my $di = ($pi >= 0 and $mi >= 0) ? (($pi < $mi) ? $pi : $mi) : (($pi >= 0) ? $pi : $mi);
        return decode_float(substr($code,0,$di)) . 'e' . substr($code,$di,1) . sprintf("%03d",decode_integer_base94(substr($code,$di+1)));
    }
    else
    {
        return decode_float($code);
    }
}
